@@ -149,6 +149,19 @@ func TestFloatingIPClientGetByNameNotFound(t *testing.T) {
 	})
 }
 
+func TestFloatingIPClientGetByNameEmpty(t *testing.T) {
+	env := newTestEnv()
+	defer env.Teardown()
+
+	floatingIP, _, err := env.Client.FloatingIP.GetByName(context.Background(), "")
+	if err != nil {
+		t.Fatal(err)
+	}
+	if floatingIP != nil {
+		t.Fatal("unexpected Floating IP")
+	}
+}
+
 func TestFloatingIPClientList(t *testing.T) {
 	env := newTestEnv()
 	defer env.Teardown()
@@ -304,9 +317,7 @@ func TestFloatingIPClientDelete(t *testing.T) {
 	env := newTestEnv()
 	defer env.Teardown()
 
-	env.Mux.HandleFunc("/floating_ips/1", func(w http.ResponseWriter, r *http.Request) {
-		return
-	})
+	env.Mux.HandleFunc("/floating_ips/1", func(w http.ResponseWriter, r *http.Request) {})
 
 	var (
 		ctx        = context.Background()
@@ -524,75 +535,6 @@ func TestFloatingIPClientUnassign(t *testing.T) {
 	}
 }
 
-func TestFloatingIPClientChangeDNSPtr(t *testing.T) {
-	var (
-		ctx        = context.Background()
-		floatingIP = &FloatingIP{ID: 1}
-	)
-
-	t.Run("set", func(t *testing.T) {
-		env := newTestEnv()
-		defer env.Teardown()
-
-		env.Mux.HandleFunc("/floating_ips/1/actions/change_dns_ptr", func(w http.ResponseWriter, r *http.Request) {
-			var reqBody schema.FloatingIPActionChangeDNSPtrRequest
-			if err := json.NewDecoder(r.Body).Decode(&reqBody); err != nil {
-				t.Fatal(err)
-			}
-			if reqBody.IP != "127.0.0.1" {
-				t.Errorf("unexpected IP: %v", reqBody.IP)
-			}
-			if reqBody.DNSPtr == nil || *reqBody.DNSPtr != "example.com" {
-				t.Errorf("unexpected DNS ptr: %v", reqBody.DNSPtr)
-			}
-			json.NewEncoder(w).Encode(schema.FloatingIPActionChangeDNSPtrResponse{
-				Action: schema.Action{
-					ID: 1,
-				},
-			})
-		})
-
-		action, _, err := env.Client.FloatingIP.ChangeDNSPtr(ctx, floatingIP, "127.0.0.1", String("example.com"))
-		if err != nil {
-			t.Fatal(err)
-		}
-		if action.ID != 1 {
-			t.Errorf("unexpected action ID: %d", action.ID)
-		}
-	})
-
-	t.Run("reset", func(t *testing.T) {
-		env := newTestEnv()
-		defer env.Teardown()
-
-		env.Mux.HandleFunc("/floating_ips/1/actions/change_dns_ptr", func(w http.ResponseWriter, r *http.Request) {
-			var reqBody schema.FloatingIPActionChangeDNSPtrRequest
-			if err := json.NewDecoder(r.Body).Decode(&reqBody); err != nil {
-				t.Fatal(err)
-			}
-			if reqBody.IP != "127.0.0.1" {
-				t.Errorf("unexpected IP: %v", reqBody.IP)
-			}
-			if reqBody.DNSPtr != nil {
-				t.Errorf("unexpected DNS ptr: %v", reqBody.DNSPtr)
-			}
-			json.NewEncoder(w).Encode(schema.FloatingIPActionChangeDNSPtrResponse{
-				Action: schema.Action{
-					ID: 1,
-				},
-			})
-		})
-
-		action, _, err := env.Client.FloatingIP.ChangeDNSPtr(ctx, floatingIP, "127.0.0.1", nil)
-		if err != nil {
-			t.Fatal(err)
-		}
-		if action.ID != 1 {
-			t.Errorf("unexpected action ID: %d", action.ID)
-		}
-	})
-}
-
 func TestFloatingIPClientChangeProtection(t *testing.T) {
 	var (
 		ctx        = context.Background()
