@@ -139,6 +139,20 @@ func TestNetworkClientGetByNameNotFound(t *testing.T) {
 	}
 }
 
+func TestNetworkClientGetByNameEmpty(t *testing.T) {
+	env := newTestEnv()
+	defer env.Teardown()
+
+	ctx := context.Background()
+	network, _, err := env.Client.Network.GetByName(ctx, "")
+	if err != nil {
+		t.Fatal(err)
+	}
+	if network != nil {
+		t.Fatal("unexpected network")
+	}
+}
+
 func TestNetworkCreate(t *testing.T) {
 	var (
 		ctx           = context.Background()
@@ -205,9 +219,7 @@ func TestNetworkDelete(t *testing.T) {
 	env := newTestEnv()
 	defer env.Teardown()
 
-	env.Mux.HandleFunc("/networks/1", func(w http.ResponseWriter, r *http.Request) {
-		return
-	})
+	env.Mux.HandleFunc("/networks/1", func(w http.ResponseWriter, r *http.Request) {})
 
 	var (
 		ctx     = context.Background()
@@ -372,7 +384,7 @@ func TestNetworkClientAddSubnet(t *testing.T) {
 			if err := json.NewDecoder(r.Body).Decode(&reqBody); err != nil {
 				t.Fatal(err)
 			}
-			if reqBody.Type != "server" {
+			if reqBody.Type != "cloud" {
 				t.Errorf("unexpected Type: %v", reqBody.Type)
 			}
 			if reqBody.IPRange != "10.0.1.0/24" {
@@ -392,7 +404,7 @@ func TestNetworkClientAddSubnet(t *testing.T) {
 		_, ipRange, _ := net.ParseCIDR("10.0.1.0/24")
 		opts := NetworkAddSubnetOpts{
 			Subnet: NetworkSubnet{
-				Type:        NetworkSubnetTypeServer,
+				Type:        NetworkSubnetTypeCloud,
 				IPRange:     ipRange,
 				NetworkZone: NetworkZoneEUCentral,
 			},
@@ -415,7 +427,7 @@ func TestNetworkClientAddSubnet(t *testing.T) {
 			if err := json.NewDecoder(r.Body).Decode(&reqBody); err != nil {
 				t.Fatal(err)
 			}
-			if reqBody.Type != "server" {
+			if reqBody.Type != "cloud" {
 				t.Errorf("unexpected Type: %v", reqBody.Type)
 			}
 			if reqBody.IPRange != "" {
@@ -434,7 +446,7 @@ func TestNetworkClientAddSubnet(t *testing.T) {
 		ctx := context.Background()
 		opts := NetworkAddSubnetOpts{
 			Subnet: NetworkSubnet{
-				Type:        NetworkSubnetTypeServer,
+				Type:        NetworkSubnetTypeCloud,
 				NetworkZone: NetworkZoneEUCentral,
 			},
 		}
@@ -446,6 +458,53 @@ func TestNetworkClientAddSubnet(t *testing.T) {
 			t.Errorf("unexpected action ID: %d", action.ID)
 		}
 	})
+
+	t.Run("type vswitch with ip range", func(t *testing.T) {
+		env := newTestEnv()
+		defer env.Teardown()
+
+		env.Mux.HandleFunc("/networks/1/actions/add_subnet", func(w http.ResponseWriter, r *http.Request) {
+			var reqBody schema.NetworkActionAddSubnetRequest
+			if err := json.NewDecoder(r.Body).Decode(&reqBody); err != nil {
+				t.Fatal(err)
+			}
+			if reqBody.Type != "vswitch" {
+				t.Errorf("unexpected Type: %v", reqBody.Type)
+			}
+			if reqBody.IPRange != "10.0.1.0/24" {
+				t.Errorf("unexpected IPRange: %v", reqBody.IPRange)
+			}
+			if reqBody.NetworkZone != "eu-central" {
+				t.Errorf("unexpected NetworkZone: %v", reqBody.NetworkZone)
+			}
+			if reqBody.VSwitchID != 123 {
+				t.Errorf("unexpected VSwitchID: %v", reqBody.VSwitchID)
+			}
+			json.NewEncoder(w).Encode(schema.NetworkActionAddSubnetResponse{
+				Action: schema.Action{
+					ID: 1,
+				},
+			})
+		})
+
+		ctx := context.Background()
+		_, ipRange, _ := net.ParseCIDR("10.0.1.0/24")
+		opts := NetworkAddSubnetOpts{
+			Subnet: NetworkSubnet{
+				Type:        NetworkSubnetTypeVSwitch,
+				IPRange:     ipRange,
+				NetworkZone: NetworkZoneEUCentral,
+				VSwitchID:   123,
+			},
+		}
+		action, _, err := env.Client.Network.AddSubnet(ctx, &Network{ID: 1}, opts)
+		if err != nil {
+			t.Fatal(err)
+		}
+		if action.ID != 1 {
+			t.Errorf("unexpected action ID: %d", action.ID)
+		}
+	})
 }
 
 func TestNetworkClientDeleteSubnet(t *testing.T) {
