@@ -23,6 +23,8 @@ type Server struct {
 	Protection      ServerProtection   `json:"protection"`
 	Labels          map[string]string  `json:"labels"`
 	Volumes         []int              `json:"volumes"`
+	PrimaryDiskSize int                `json:"primary_disk_size"`
+	PlacementGroup  *PlacementGroup    `json:"placement_group"`
 }
 
 // ServerProtection defines the schema of a server's resource protection.
@@ -37,11 +39,13 @@ type ServerPublicNet struct {
 	IPv4        ServerPublicNetIPv4 `json:"ipv4"`
 	IPv6        ServerPublicNetIPv6 `json:"ipv6"`
 	FloatingIPs []int               `json:"floating_ips"`
+	Firewalls   []ServerFirewall    `json:"firewalls"`
 }
 
 // ServerPublicNetIPv4 defines the schema of a server's public
 // network information for an IPv4.
 type ServerPublicNetIPv4 struct {
+	ID      int    `json:"id"`
 	IP      string `json:"ip"`
 	Blocked bool   `json:"blocked"`
 	DNSPtr  string `json:"dns_ptr"`
@@ -50,6 +54,7 @@ type ServerPublicNetIPv4 struct {
 // ServerPublicNetIPv6 defines the schema of a server's public
 // network information for an IPv6.
 type ServerPublicNetIPv6 struct {
+	ID      int                         `json:"id"`
 	IP      string                      `json:"ip"`
 	Blocked bool                        `json:"blocked"`
 	DNSPtr  []ServerPublicNetIPv6DNSPtr `json:"dns_ptr"`
@@ -62,6 +67,13 @@ type ServerPublicNetIPv6DNSPtr struct {
 	DNSPtr string `json:"dns_ptr"`
 }
 
+// ServerFirewall defines the schema of a Server's Firewalls on
+// a certain network interface.
+type ServerFirewall struct {
+	ID     int    `json:"id"`
+	Status string `json:"status"`
+}
+
 // ServerPrivateNet defines the schema of a server's private network information.
 type ServerPrivateNet struct {
 	Network    int      `json:"network"`
@@ -85,18 +97,34 @@ type ServerListResponse struct {
 // ServerCreateRequest defines the schema for the request to
 // create a server.
 type ServerCreateRequest struct {
-	Name             string             `json:"name"`
-	ServerType       interface{}        `json:"server_type"` // int or string
-	Image            interface{}        `json:"image"`       // int or string
-	SSHKeys          []int              `json:"ssh_keys,omitempty"`
-	Location         string             `json:"location,omitempty"`
-	Datacenter       string             `json:"datacenter,omitempty"`
-	UserData         string             `json:"user_data,omitempty"`
-	StartAfterCreate *bool              `json:"start_after_create,omitempty"`
-	Labels           *map[string]string `json:"labels,omitempty"`
-	Automount        *bool              `json:"automount,omitempty"`
-	Volumes          []int              `json:"volumes,omitempty"`
-	Networks         []int              `json:"networks,omitempty"`
+	Name             string                  `json:"name"`
+	ServerType       interface{}             `json:"server_type"` // int or string
+	Image            interface{}             `json:"image"`       // int or string
+	SSHKeys          []int                   `json:"ssh_keys,omitempty"`
+	Location         string                  `json:"location,omitempty"`
+	Datacenter       string                  `json:"datacenter,omitempty"`
+	UserData         string                  `json:"user_data,omitempty"`
+	StartAfterCreate *bool                   `json:"start_after_create,omitempty"`
+	Labels           *map[string]string      `json:"labels,omitempty"`
+	Automount        *bool                   `json:"automount,omitempty"`
+	Volumes          []int                   `json:"volumes,omitempty"`
+	Networks         []int                   `json:"networks,omitempty"`
+	Firewalls        []ServerCreateFirewalls `json:"firewalls,omitempty"`
+	PlacementGroup   int                     `json:"placement_group,omitempty"`
+	PublicNet        *ServerCreatePublicNet  `json:"public_net,omitempty"`
+}
+
+// ServerCreatePublicNet defines the public network configuration of a server.
+type ServerCreatePublicNet struct {
+	EnableIPv4 bool `json:"enable_ipv4"`
+	EnableIPv6 bool `json:"enable_ipv6"`
+	IPv4ID     int  `json:"ipv4,omitempty"`
+	IPv6ID     int  `json:"ipv6,omitempty"`
+}
+
+// ServerCreateFirewalls defines which Firewalls to apply when creating a Server.
+type ServerCreateFirewalls struct {
+	Firewall int `json:"firewall"`
 }
 
 // ServerCreateResponse defines the schema of the response when
@@ -314,6 +342,18 @@ type ServerActionChangeProtectionResponse struct {
 	Action Action `json:"action"`
 }
 
+// ServerActionRequestConsoleRequest defines the schema of the request to
+// request a WebSocket VNC console.
+type ServerActionRequestConsoleRequest struct{}
+
+// ServerActionRequestConsoleResponse defines the schema of the response when
+// requesting a WebSocket VNC console.
+type ServerActionRequestConsoleResponse struct {
+	Action   Action `json:"action"`
+	WSSURL   string `json:"wss_url"`
+	Password string `json:"password"`
+}
+
 // ServerActionAttachToNetworkRequest defines the schema for the request to
 // attach a network to a server.
 type ServerActionAttachToNetworkRequest struct {
@@ -352,3 +392,37 @@ type ServerActionChangeAliasIPsRequest struct {
 type ServerActionChangeAliasIPsResponse struct {
 	Action Action `json:"action"`
 }
+
+// ServerGetMetricsResponse defines the schema of the response when requesting
+// metrics for a server.
+type ServerGetMetricsResponse struct {
+	Metrics struct {
+		Start      time.Time                       `json:"start"`
+		End        time.Time                       `json:"end"`
+		Step       float64                         `json:"step"`
+		TimeSeries map[string]ServerTimeSeriesVals `json:"time_series"`
+	} `json:"metrics"`
+}
+
+// ServerTimeSeriesVals contains the values for a Server time series.
+type ServerTimeSeriesVals struct {
+	Values []interface{} `json:"values"`
+}
+
+// ServerActionAddToPlacementGroupRequest defines the schema for the request to
+// add a server to a placement group.
+type ServerActionAddToPlacementGroupRequest struct {
+	PlacementGroup int `json:"placement_group"`
+}
+
+// ServerActionAddToPlacementGroupResponse defines the schema of the response when
+// creating an add_to_placement_group server action.
+type ServerActionAddToPlacementGroupResponse struct {
+	Action Action `json:"action"`
+}
+
+// ServerActionRemoveFromPlacementGroupResponse defines the schema of the response when
+// creating a remove_from_placement_group server action.
+type ServerActionRemoveFromPlacementGroupResponse struct {
+	Action Action `json:"action"`
+}
