@@ -3,10 +3,15 @@ package hcloud
 import (
 	"context"
 	"encoding/json"
+	"fmt"
 	"net"
 	"net/http"
+	"net/url"
+	"strconv"
 	"testing"
+	"time"
 
+	"github.com/google/go-cmp/cmp"
 	"github.com/hetznercloud/hcloud-go/hcloud/schema"
 )
 
@@ -139,6 +144,20 @@ func TestServerClientGetByNameNotFound(t *testing.T) {
 	}
 }
 
+func TestServerClientGetByNameEmpty(t *testing.T) {
+	env := newTestEnv()
+	defer env.Teardown()
+
+	ctx := context.Background()
+	server, _, err := env.Client.Server.GetByName(ctx, "")
+	if err != nil {
+		t.Fatal(err)
+	}
+	if server != nil {
+		t.Fatal("unexpected server")
+	}
+}
+
 func TestServersList(t *testing.T) {
 	env := newTestEnv()
 	defer env.Teardown()
@@ -442,6 +461,156 @@ func TestServersCreateWithNetworks(t *testing.T) {
 	}
 }
 
+func TestServersCreateWithPrivateNetworkOnly(t *testing.T) {
+	env := newTestEnv()
+	defer env.Teardown()
+
+	env.Mux.HandleFunc("/servers", func(w http.ResponseWriter, r *http.Request) {
+		var reqBody schema.ServerCreateRequest
+		if err := json.NewDecoder(r.Body).Decode(&reqBody); err != nil {
+			t.Fatal(err)
+		}
+		if len(reqBody.Networks) != 2 || reqBody.Networks[0] != 1 || reqBody.Networks[1] != 2 {
+			t.Errorf("unexpected Networks: %v", reqBody.Networks)
+		}
+		if reqBody.PublicNet.EnableIPv4 != false {
+			t.Errorf("unexpected PublicNet.EnableIPv4: %v", reqBody.PublicNet.EnableIPv4)
+		}
+		if reqBody.PublicNet.EnableIPv6 != false {
+			t.Errorf("unexpected PublicNet.EnableIPv6: %v", reqBody.PublicNet.EnableIPv6)
+		}
+		if reqBody.PublicNet.IPv4ID != 0 {
+			t.Errorf("unexpected PublicNet.IPv4: %v", reqBody.PublicNet.IPv4ID)
+		}
+		if reqBody.PublicNet.IPv6ID != 0 {
+			t.Errorf("unexpected PublicNet.IPv6: %v", reqBody.PublicNet.IPv6ID)
+		}
+		json.NewEncoder(w).Encode(schema.ServerCreateResponse{
+			Server: schema.Server{
+				ID: 1,
+			},
+			NextActions: []schema.Action{
+				{ID: 2},
+			},
+		})
+	})
+
+	ctx := context.Background()
+	result, _, err := env.Client.Server.Create(ctx, ServerCreateOpts{
+		Name:       "test",
+		ServerType: &ServerType{ID: 1},
+		Image:      &Image{ID: 2},
+		Networks: []*Network{
+			{ID: 1},
+			{ID: 2},
+		},
+		PublicNet: &ServerCreatePublicNet{
+			EnableIPv4: false,
+			EnableIPv6: false,
+		},
+	})
+	if err != nil {
+		t.Fatal(err)
+	}
+	if result.Server == nil {
+		t.Fatal("no server")
+	}
+	if result.Server.ID != 1 {
+		t.Errorf("unexpected server ID: %v", result.Server.ID)
+	}
+	if len(result.NextActions) != 1 || result.NextActions[0].ID != 2 {
+		t.Errorf("unexpected next actions: %v", result.NextActions)
+	}
+}
+
+func TestServersCreateWithIPv6Only(t *testing.T) {
+	env := newTestEnv()
+	defer env.Teardown()
+
+	env.Mux.HandleFunc("/servers", func(w http.ResponseWriter, r *http.Request) {
+		var reqBody schema.ServerCreateRequest
+		if err := json.NewDecoder(r.Body).Decode(&reqBody); err != nil {
+			t.Fatal(err)
+		}
+		if reqBody.PublicNet.EnableIPv4 != false {
+			t.Errorf("unexpected PublicNet.EnableIPv4: %v", reqBody.PublicNet.EnableIPv4)
+		}
+		if reqBody.PublicNet.EnableIPv6 != true {
+			t.Errorf("unexpected PublicNet.EnableIPv6: %v", reqBody.PublicNet.EnableIPv6)
+		}
+		json.NewEncoder(w).Encode(schema.ServerCreateResponse{
+			Server: schema.Server{
+				ID: 1,
+			},
+			NextActions: []schema.Action{
+				{ID: 2},
+			},
+		})
+	})
+
+	ctx := context.Background()
+	result, _, err := env.Client.Server.Create(ctx, ServerCreateOpts{
+		Name:       "test",
+		ServerType: &ServerType{ID: 1},
+		Image:      &Image{ID: 2},
+		PublicNet:  &ServerCreatePublicNet{EnableIPv4: false, EnableIPv6: true},
+	})
+	if err != nil {
+		t.Fatal(err)
+	}
+	if result.Server == nil {
+		t.Fatal("no server")
+	}
+	if result.Server.ID != 1 {
+		t.Errorf("unexpected server ID: %v", result.Server.ID)
+	}
+	if len(result.NextActions) != 1 || result.NextActions[0].ID != 2 {
+		t.Errorf("unexpected next actions: %v", result.NextActions)
+	}
+}
+
+func TestServersCreateWithDefaultPublicNet(t *testing.T) {
+	env := newTestEnv()
+	defer env.Teardown()
+
+	env.Mux.HandleFunc("/servers", func(w http.ResponseWriter, r *http.Request) {
+		var reqBody schema.ServerCreateRequest
+		if err := json.NewDecoder(r.Body).Decode(&reqBody); err != nil {
+			t.Fatal(err)
+		}
+		if reqBody.PublicNet != nil {
+			t.Errorf("unexpected PublicNet: %v", reqBody.PublicNet)
+		}
+		json.NewEncoder(w).Encode(schema.ServerCreateResponse{
+			Server: schema.Server{
+				ID: 1,
+			},
+			NextActions: []schema.Action{
+				{ID: 2},
+			},
+		})
+	})
+
+	ctx := context.Background()
+	result, _, err := env.Client.Server.Create(ctx, ServerCreateOpts{
+		Name:       "test",
+		ServerType: &ServerType{ID: 1},
+		Image:      &Image{ID: 2},
+	})
+	if err != nil {
+		t.Fatal(err)
+	}
+	if result.Server == nil {
+		t.Fatal("no server")
+	}
+	if result.Server.ID != 1 {
+		t.Errorf("unexpected server ID: %v", result.Server.ID)
+	}
+	if len(result.NextActions) != 1 || result.NextActions[0].ID != 2 {
+		t.Errorf("unexpected next actions: %v", result.NextActions)
+	}
+}
+
 func TestServersCreateWithDatacenterID(t *testing.T) {
 	env := newTestEnv()
 	defer env.Teardown()
@@ -612,6 +781,52 @@ func TestServersCreateWithUserData(t *testing.T) {
 	}
 }
 
+func TestServersCreateWithFirewalls(t *testing.T) {
+	env := newTestEnv()
+	defer env.Teardown()
+
+	env.Mux.HandleFunc("/servers", func(w http.ResponseWriter, r *http.Request) {
+		var reqBody schema.ServerCreateRequest
+		if err := json.NewDecoder(r.Body).Decode(&reqBody); err != nil {
+			t.Fatal(err)
+		}
+		if len(reqBody.Firewalls) != 2 || reqBody.Firewalls[0].Firewall != 1 || reqBody.Firewalls[1].Firewall != 2 {
+			t.Errorf("unexpected Firewalls: %v", reqBody.Firewalls)
+		}
+		json.NewEncoder(w).Encode(schema.ServerCreateResponse{
+			Server: schema.Server{
+				ID: 1,
+			},
+			NextActions: []schema.Action{
+				{ID: 2},
+			},
+		})
+	})
+
+	ctx := context.Background()
+	result, _, err := env.Client.Server.Create(ctx, ServerCreateOpts{
+		Name:       "test",
+		ServerType: &ServerType{ID: 1},
+		Image:      &Image{ID: 2},
+		Firewalls: []*ServerCreateFirewall{
+			{Firewall: Firewall{ID: 1}},
+			{Firewall: Firewall{ID: 2}},
+		},
+	})
+	if err != nil {
+		t.Fatal(err)
+	}
+	if result.Server == nil {
+		t.Fatal("no server")
+	}
+	if result.Server.ID != 1 {
+		t.Errorf("unexpected server ID: %v", result.Server.ID)
+	}
+	if len(result.NextActions) != 1 || result.NextActions[0].ID != 2 {
+		t.Errorf("unexpected next actions: %v", result.NextActions)
+	}
+}
+
 func TestServersCreateWithLabels(t *testing.T) {
 	env := newTestEnv()
 	defer env.Teardown()
@@ -687,13 +902,127 @@ func TestServersCreateWithoutStarting(t *testing.T) {
 	}
 }
 
+func TestServersCreateFailIfNoIPsAndNetworksAssigned(t *testing.T) {
+	env := newTestEnv()
+	defer env.Teardown()
+	ctx := context.Background()
+	_, _, err := env.Client.Server.Create(ctx, ServerCreateOpts{
+		Name:       "test",
+		ServerType: &ServerType{ID: 1},
+		Image:      &Image{ID: 2},
+		PublicNet: &ServerCreatePublicNet{
+			EnableIPv4: false,
+			EnableIPv6: false,
+		},
+	})
+	if err == nil {
+		t.Fatal(err)
+	}
+}
+
+func TestServerCreateWithPlacementGroup(t *testing.T) {
+	env := newTestEnv()
+	defer env.Teardown()
+
+	env.Mux.HandleFunc("/servers", func(w http.ResponseWriter, r *http.Request) {
+		var reqBody schema.ServerCreateRequest
+		if err := json.NewDecoder(r.Body).Decode(&reqBody); err != nil {
+			t.Fatal(err)
+		}
+		if reqBody.PlacementGroup != 123 {
+			t.Errorf("unexpected placement group id %d", reqBody.PlacementGroup)
+		}
+		json.NewEncoder(w).Encode(schema.ServerCreateResponse{
+			Server: schema.Server{
+				ID: 1,
+				PlacementGroup: &schema.PlacementGroup{
+					ID: 123,
+				},
+			},
+			NextActions: []schema.Action{
+				{ID: 2},
+			},
+		})
+	})
+
+	ctx := context.Background()
+	result, _, err := env.Client.Server.Create(ctx, ServerCreateOpts{
+		Name:           "test",
+		ServerType:     &ServerType{ID: 1},
+		Image:          &Image{ID: 2},
+		PlacementGroup: &PlacementGroup{ID: 123},
+	})
+	if err != nil {
+		t.Fatal(err)
+	}
+	if result.Server == nil {
+		t.Fatal("no server")
+	}
+	if result.Server.ID != 1 {
+		t.Errorf("unexpected server ID: %d", result.Server.ID)
+	}
+	if len(result.NextActions) != 1 || result.NextActions[0].ID != 2 {
+		t.Errorf("unexpected next actions: %v", result.NextActions)
+	}
+	if result.Server.PlacementGroup.ID != 123 {
+		t.Errorf("unexpected placement group ID: %d", result.Server.PlacementGroup.ID)
+	}
+}
+
+func TestServerCreateWithoutPrimaryIPsButNetwork(t *testing.T) {
+	env := newTestEnv()
+	defer env.Teardown()
+
+	env.Mux.HandleFunc("/servers", func(w http.ResponseWriter, r *http.Request) {
+		var reqBody schema.ServerCreateRequest
+		if err := json.NewDecoder(r.Body).Decode(&reqBody); err != nil {
+			t.Fatal(err)
+		}
+		if reqBody.PublicNet.EnableIPv4 != false && reqBody.PublicNet.EnableIPv6 != false {
+			t.Errorf("unexpected public net %v", reqBody.PublicNet)
+		}
+		json.NewEncoder(w).Encode(schema.ServerCreateResponse{
+			Server: schema.Server{
+				ID: 1,
+			},
+			NextActions: []schema.Action{
+				{ID: 2},
+			},
+		})
+	})
+
+	ctx := context.Background()
+	result, _, err := env.Client.Server.Create(ctx, ServerCreateOpts{
+		Name:       "test",
+		ServerType: &ServerType{ID: 1},
+		Image:      &Image{ID: 2},
+		Networks: []*Network{
+			{ID: 1},
+		},
+		PublicNet: &ServerCreatePublicNet{
+			EnableIPv4: false,
+			EnableIPv6: false,
+		},
+	})
+	if err != nil {
+		t.Fatal(err)
+	}
+	if result.Server == nil {
+		t.Fatal("no server")
+	}
+	if result.Server.ID != 1 {
+		t.Errorf("unexpected server ID: %d", result.Server.ID)
+	}
+	if len(result.NextActions) != 1 || result.NextActions[0].ID != 2 {
+		t.Errorf("unexpected next actions: %v", result.NextActions)
+	}
+}
+
 func TestServersDelete(t *testing.T) {
 	env := newTestEnv()
 	defer env.Teardown()
 
-	env.Mux.HandleFunc("/servers/1", func(w http.ResponseWriter, r *http.Request) {
-		return
-	})
+	env.Mux.HandleFunc("/servers/1", func(w http.ResponseWriter, r *http.Request) {})
 
 	var (
 		ctx    = context.Background()
@@ -744,7 +1073,6 @@ func TestServerClientUpdate(t *testing.T) {
 		if updatedServer.ID != 1 {
 			t.Errorf("unexpected server ID: %v", updatedServer.ID)
 		}
-
 	})
 
 	t.Run("update labels", func(t *testing.T) {
@@ -1413,75 +1741,6 @@ func TestServerClientChangeType(t *testing.T) {
 	})
 }
 
-func TestServerClientChangeDNSPtr(t *testing.T) {
-	var (
-		ctx    = context.Background()
-		server = &Server{ID: 1}
-	)
-
-	t.Run("set", func(t *testing.T) {
-		env := newTestEnv()
-		defer env.Teardown()
-
-		env.Mux.HandleFunc("/servers/1/actions/change_dns_ptr", func(w http.ResponseWriter, r *http.Request) {
-			var reqBody schema.ServerActionChangeDNSPtrRequest
-			if err := json.NewDecoder(r.Body).Decode(&reqBody); err != nil {
-				t.Fatal(err)
-			}
-			if reqBody.IP != "127.0.0.1" {
-				t.Errorf("unexpected IP: %v", reqBody.IP)
-			}
-			if reqBody.DNSPtr == nil || *reqBody.DNSPtr != "example.com" {
-				t.Errorf("unexpected DNS ptr: %v", reqBody.DNSPtr)
-			}
-			json.NewEncoder(w).Encode(schema.ServerActionChangeDNSPtrResponse{
-				Action: schema.Action{
-					ID: 1,
-				},
-			})
-		})
-
-		action, _, err := env.Client.Server.ChangeDNSPtr(ctx, server, "127.0.0.1", String("example.com"))
-		if err != nil {
-			t.Fatal(err)
-		}
-		if action.ID != 1 {
-			t.Errorf("unexpected action ID: %d", action.ID)
-		}
-	})
-
-	t.Run("reset", func(t *testing.T) {
-		env := newTestEnv()
-		defer env.Teardown()
-
-		env.Mux.HandleFunc("/servers/1/actions/change_dns_ptr", func(w http.ResponseWriter, r *http.Request) {
-			var reqBody schema.ServerActionChangeDNSPtrRequest
-			if err := json.NewDecoder(r.Body).Decode(&reqBody); err != nil {
-				t.Fatal(err)
-			}
-			if reqBody.IP != "127.0.0.1" {
-				t.Errorf("unexpected IP: %v", reqBody.IP)
-			}
-			if reqBody.DNSPtr != nil {
-				t.Errorf("unexpected DNS ptr: %v", reqBody.DNSPtr)
-			}
-			json.NewEncoder(w).Encode(schema.ServerActionChangeDNSPtrResponse{
-				Action: schema.Action{
-					ID: 1,
-				},
-			})
-		})
-
-		action, _, err := env.Client.Server.ChangeDNSPtr(ctx, server, "127.0.0.1", nil)
-		if err != nil {
-			t.Fatal(err)
-		}
-		if action.ID != 1 {
-			t.Errorf("unexpected action ID: %d", action.ID)
-		}
-	})
-}
-
 func TestServerClientChangeProtection(t *testing.T) {
 	var (
 		ctx    = context.Background()
@@ -1528,6 +1787,36 @@ func TestServerClientChangeProtection(t *testing.T) {
 	})
 }
 
+func TestServerClientRequestConsole(t *testing.T) {
+	env := newTestEnv()
+	defer env.Teardown()
+
+	env.Mux.HandleFunc("/servers/1/actions/request_console", func(w http.ResponseWriter, r *http.Request) {
+		json.NewEncoder(w).Encode(schema.ServerActionRequestConsoleResponse{
+			Action: schema.Action{
+				ID: 1,
+			},
+			WSSURL:   "wss://console.hetzner.cloud/?server_id=1&token=3db32d15-af2f-459c-8bf8-dee1fd05f49c",
+			Password: "9MQaTg2VAGI0FIpc10k3UpRXcHj2wQ6x",
+		})
+	})
+
+	ctx := context.Background()
+	result, _, err := env.Client.Server.RequestConsole(ctx, &Server{ID: 1})
+	if err != nil {
+		t.Fatal(err)
+	}
+	if result.Action.ID != 1 {
+		t.Errorf("unexpected action ID: %d", result.Action.ID)
+	}
+	if result.WSSURL != "wss://console.hetzner.cloud/?server_id=1&token=3db32d15-af2f-459c-8bf8-dee1fd05f49c" {
+		t.Errorf("unexpected WebSocket URL: %v", result.WSSURL)
+	}
+	if result.Password != "9MQaTg2VAGI0FIpc10k3UpRXcHj2wQ6x" {
+		t.Errorf("unexpected password: %v", result.Password)
+	}
+}
+
 func TestServerClientAttachToNetwork(t *testing.T) {
 	var (
 		ctx    = context.Background()
@@ -1702,3 +1991,348 @@ func TestServerClientChangeAliasIP(t *testing.T) {
 		t.Errorf("unexpected action ID: %v", action.ID)
 	}
 }
+
+func TestServerGetMetrics(t *testing.T) {
+	tests := []struct {
+		name        string
+		server      *Server
+		opts        ServerGetMetricsOpts
+		respStatus  int
+		respFn      func() schema.ServerGetMetricsResponse
+		expected    ServerMetrics
+		expectedErr string
+	}{
+		{
+			name:   "cpu metrics",
+			server: &Server{ID: 1},
+			opts: ServerGetMetricsOpts{
+				Types: []ServerMetricType{ServerMetricCPU},
+				Start: mustParseTime(t, time.RFC3339, "2017-01-01T00:00:00Z"),
+				End:   mustParseTime(t, time.RFC3339, "2017-01-01T23:00:00Z"),
+			},
+			respFn: func() schema.ServerGetMetricsResponse {
+				var resp schema.ServerGetMetricsResponse
+
+				resp.Metrics.Start = mustParseTime(t, time.RFC3339, "2017-01-01T00:00:00Z")
+				resp.Metrics.End = mustParseTime(t, time.RFC3339, "2017-01-01T23:00:00Z")
+				resp.Metrics.TimeSeries = map[string]schema.ServerTimeSeriesVals{
+					"cpu": {
+						Values: []interface{}{
+							[]interface{}{1435781470.622, "42"},
+							[]interface{}{1435781471.622, "43"},
+						},
+					},
+				}
+
+				return resp
+			},
+			expected: ServerMetrics{
+				Start: mustParseTime(t, time.RFC3339, "2017-01-01T00:00:00Z"),
+				End:   mustParseTime(t, time.RFC3339, "2017-01-01T23:00:00Z"),
+				TimeSeries: map[string][]ServerMetricsValue{
+					"cpu": {
+						{Timestamp: 1435781470.622, Value: "42"},
+						{Timestamp: 1435781471.622, Value: "43"},
+					},
+				},
+			},
+		},
+		{
+			name:   "all metrics",
+			server: &Server{ID: 2},
+			opts: ServerGetMetricsOpts{
+				Types: []ServerMetricType{
+					ServerMetricCPU,
+					ServerMetricDisk,
+					ServerMetricNetwork,
+				},
+				Start: mustParseTime(t, time.RFC3339, "2017-01-01T00:00:00Z"),
+				End:   mustParseTime(t, time.RFC3339, "2017-01-01T23:00:00Z"),
+			},
+			respFn: func() schema.ServerGetMetricsResponse {
+				var resp schema.ServerGetMetricsResponse
+
+				resp.Metrics.Start = mustParseTime(t, time.RFC3339, "2017-01-01T00:00:00Z")
+				resp.Metrics.End = mustParseTime(t, time.RFC3339, "2017-01-01T23:00:00Z")
+				resp.Metrics.TimeSeries = map[string]schema.ServerTimeSeriesVals{
+					"cpu": {
+						Values: []interface{}{
+							[]interface{}{1435781470.622, "42"},
+							[]interface{}{1435781471.622, "43"},
+						},
+					},
+					"disk.0.iops.read": {
+						Values: []interface{}{
+							[]interface{}{1435781480.622, "100"},
+							[]interface{}{1435781481.622, "150"},
+						},
+					},
+					"disk.0.iops.write": {
+						Values: []interface{}{
+							[]interface{}{1435781480.622, "50"},
+							[]interface{}{1435781481.622, "55"},
+						},
+					},
+					"network.0.pps.in": {
+						Values: []interface{}{
+							[]interface{}{1435781490.622, "70"},
+							[]interface{}{1435781491.622, "75"},
+						},
+					},
+					"network.0.pps.out": {
+						Values: []interface{}{
+							[]interface{}{1435781590.622, "60"},
+							[]interface{}{1435781591.622, "65"},
+						},
+					},
+				}
+
+				return resp
+			},
+			expected: ServerMetrics{
+				Start: mustParseTime(t, time.RFC3339, "2017-01-01T00:00:00Z"),
+				End:   mustParseTime(t, time.RFC3339, "2017-01-01T23:00:00Z"),
+				TimeSeries: map[string][]ServerMetricsValue{
+					"cpu": {
+						{Timestamp: 1435781470.622, Value: "42"},
+						{Timestamp: 1435781471.622, Value: "43"},
+					},
+					"disk.0.iops.read": {
+						{Timestamp: 1435781480.622, Value: "100"},
+						{Timestamp: 1435781481.622, Value: "150"},
+					},
+					"disk.0.iops.write": {
+						{Timestamp: 1435781480.622, Value: "50"},
+						{Timestamp: 1435781481.622, Value: "55"},
+					},
+					"network.0.pps.in": {
+						{Timestamp: 1435781490.622, Value: "70"},
+						{Timestamp: 1435781491.622, Value: "75"},
+					},
+					"network.0.pps.out": {
+						{Timestamp: 1435781590.622, Value: "60"},
+						{Timestamp: 1435781591.622, Value: "65"},
+					},
+				},
+			},
+		},
+		{
+			name:   "missing metrics types",
+			server: &Server{ID: 3},
+			opts: ServerGetMetricsOpts{
+				Start: mustParseTime(t, time.RFC3339, "2017-01-01T00:00:00Z"),
+				End:   mustParseTime(t, time.RFC3339, "2017-01-01T23:00:00Z"),
+			},
+			expectedErr: "add query params: no metric types specified",
+		},
+		{
+			name:   "no start time",
+			server: &Server{ID: 4},
+			opts: ServerGetMetricsOpts{
+				Types: []ServerMetricType{ServerMetricCPU},
+				End:   mustParseTime(t, time.RFC3339, "2017-01-01T23:00:00Z"),
+			},
+			expectedErr: "add query params: no start time specified",
+		},
+		{
+			name:   "no end time",
+			server: &Server{ID: 5},
+			opts: ServerGetMetricsOpts{
+				Types: []ServerMetricType{ServerMetricCPU},
+				Start: mustParseTime(t, time.RFC3339, "2017-01-01T00:00:00Z"),
+			},
+			expectedErr: "add query params: no end time specified",
+		},
+		{
+			name:   "call to backend API fails",
+			server: &Server{ID: 6},
+			opts: ServerGetMetricsOpts{
+				Types: []ServerMetricType{ServerMetricCPU},
+				Start: mustParseTime(t, time.RFC3339, "2017-01-01T00:00:00Z"),
+				End:   mustParseTime(t, time.RFC3339, "2017-01-01T23:00:00Z"),
+			},
+			respStatus:  http.StatusInternalServerError,
+			expectedErr: "get metrics: hcloud: server responded with status code 500",
+		},
+		{
+			name: "no server passed",
+			opts: ServerGetMetricsOpts{
+				Types: []ServerMetricType{ServerMetricCPU},
+				Start: mustParseTime(t, time.RFC3339, "2017-01-01T00:00:00Z"),
+				End:   mustParseTime(t, time.RFC3339, "2017-01-01T23:00:00Z"),
+			},
+			expectedErr: "illegal argument: server is nil",
+		},
+	}
+
+	for _, tt := range tests {
+		tt := tt
+		t.Run(tt.name, func(t *testing.T) {
+			env := newTestEnv()
+			defer env.Teardown()
+
+			if tt.server != nil {
+				path := fmt.Sprintf("/servers/%d/metrics", tt.server.ID)
+				env.Mux.HandleFunc(path, func(rw http.ResponseWriter, r *http.Request) {
+					if r.Method != "GET" {
+						t.Errorf("expected GET; got %s", r.Method)
+					}
+					opts := serverMetricsOptsFromURL(t, r.URL)
+					if !cmp.Equal(tt.opts, opts) {
+						t.Errorf("unexpected opts: url: %s\n%v", r.URL.String(), cmp.Diff(tt.opts, opts))
+					}
+
+					status := tt.respStatus
+					if status == 0 {
+						status = http.StatusOK
+					}
+					rw.WriteHeader(status)
+
+					if tt.respFn != nil {
+						resp := tt.respFn()
+						if err := json.NewEncoder(rw).Encode(resp); err != nil {
+							t.Errorf("failed to encode response: %v", err)
+						}
+					}
+				})
+			}
+
+			ctx := context.Background()
+			actual, _, err := env.Client.Server.GetMetrics(ctx, tt.server, tt.opts)
+			if tt.expectedErr != "" {
+				if tt.expectedErr != err.Error() {
+					t.Errorf("expected err: %v; got: %v", tt.expectedErr, err)
+				}
+				return
+			}
+			if err != nil {
+				t.Fatalf("failed to get server metrics: %v", err)
+			}
+			if !cmp.Equal(&tt.expected, actual) {
+				t.Errorf("Actual metrics did not equal expected: %s", cmp.Diff(&tt.expected, actual))
+			}
+		})
+	}
+}
+
+func serverMetricsOptsFromURL(t *testing.T, u *url.URL) ServerGetMetricsOpts {
+	var opts ServerGetMetricsOpts
+
+	for k, vs := range u.Query() {
+		switch k {
+		case "type":
+			for _, v := range vs {
+				opts.Types = append(opts.Types, ServerMetricType(v))
+			}
+		case "start":
+			if len(vs) != 1 {
+				t.Errorf("expected one value for start; got %d: %v", len(vs), vs)
+				continue
+			}
+			v, err := time.Parse(time.RFC3339, vs[0])
+			if err != nil {
+				t.Errorf("parse start as RFC3339: %v", err)
+			}
+			opts.Start = v
+		case "end":
+			if len(vs) != 1 {
+				t.Errorf("expected one value for end; got %d: %v", len(vs), vs)
+				continue
+			}
+			v, err := time.Parse(time.RFC3339, vs[0])
+			if err != nil {
+				t.Errorf("parse end as RFC3339: %v", err)
+			}
+			opts.End = v
+		case "step":
+			if len(vs) != 1 {
+				t.Errorf("expected one value for step; got %d: %v", len(vs), vs)
+				continue
+			}
+			v, err := strconv.Atoi(vs[0])
+			if err != nil {
+				t.Errorf("invalid step: %v", err)
+			}
+			opts.Step = v
+		}
+	}
+
+	return opts
+}
+
+func TestServerAddToPlacementGroup(t *testing.T) {
+	env := newTestEnv()
+	defer env.Teardown()
+
+	const (
+		serverID         = 1
+		actionID         = 42
+		placementGroupID = 123
+	)
+
+	env.Mux.HandleFunc(fmt.Sprintf("/servers/%d/actions/add_to_placement_group", serverID), func(w http.ResponseWriter, r *http.Request) {
+		if r.Method != "POST" {
+			t.Error("expected POST")
+		}
+		var reqBody schema.ServerActionAddToPlacementGroupRequest
+		if err := json.NewDecoder(r.Body).Decode(&reqBody); err != nil {
+			t.Fatal(err)
+		}
+		if reqBody.PlacementGroup != placementGroupID {
+			t.Errorf("unexpected PlacementGroup: %v", reqBody.PlacementGroup)
+		}
+		json.NewEncoder(w).Encode(schema.ServerActionAddToPlacementGroupResponse{
+			Action: schema.Action{
+				ID: actionID,
+			},
+		})
+	})
+
+	var (
+		ctx            = context.Background()
+		server         = &Server{ID: serverID}
+		placementGroup = &PlacementGroup{ID: placementGroupID}
+	)
+
+	action, _, err := env.Client.Server.AddToPlacementGroup(ctx, server, placementGroup)
+	if err != nil {
+		t.Fatal(err)
+	}
+	if action.ID != actionID {
+		t.Errorf("unexpected action ID: %v", action.ID)
+	}
+}
+
+func TestServerRemoveFromPlacementGroup(t *testing.T) {
+	env := newTestEnv()
+	defer env.Teardown()
+
+	const (
+		serverID = 1
+		actionID = 42
+	)
+
+	env.Mux.HandleFunc(fmt.Sprintf("/servers/%d/actions/remove_from_placement_group", serverID), func(w http.ResponseWriter, r *http.Request) {
+		if r.Method != "POST" {
+			t.Error("expected POST")
+		}
+		json.NewEncoder(w).Encode(schema.ServerActionRemoveFromPlacementGroupResponse{
+			Action: schema.Action{
+				ID: actionID,
+			},
+		})
+	})
+
+	var (
+		ctx    = context.Background()
+		server = &Server{ID: serverID}
+	)
+
+	action, _, err := env.Client.Server.RemoveFromPlacementGroup(ctx, server)
+	if err != nil {
+		t.Fatal(err)
+	}
+	if action.ID != actionID {
+		t.Errorf("unexpected action ID: %v", action.ID)
+	}
+}
