@@ -48,6 +48,9 @@ func (c *SSHKeyClient) GetByID(ctx context.Context, id int) (*SSHKey, *Response,
 
 // GetByName retrieves a SSH key by its name. If the SSH key does not exist, nil is returned.
 func (c *SSHKeyClient) GetByName(ctx context.Context, name string) (*SSHKey, *Response, error) {
+	if name == "" {
+		return nil, nil, nil
+	}
 	sshKeys, response, err := c.List(ctx, SSHKeyListOpts{Name: name})
 	if len(sshKeys) == 0 {
 		return nil, response, err
@@ -78,6 +81,7 @@ type SSHKeyListOpts struct {
 	ListOpts
 	Name        string
 	Fingerprint string
+	Sort        []string
 }
 
 func (l SSHKeyListOpts) values() url.Values {
@@ -88,10 +92,16 @@ func (l SSHKeyListOpts) values() url.Values {
 	if l.Fingerprint != "" {
 		vals.Add("fingerprint", l.Fingerprint)
 	}
+	for _, sort := range l.Sort {
+		vals.Add("sort", sort)
+	}
 	return vals
 }
 
 // List returns a list of SSH keys for a specific page.
+//
+// Please note that filters specified in opts are not taken into account
+// when their value corresponds to their zero value or when they are empty.
 func (c *SSHKeyClient) List(ctx context.Context, opts SSHKeyListOpts) ([]*SSHKey, *Response, error) {
 	path := "/ssh_keys?" + opts.values().Encode()
 	req, err := c.client.NewRequest(ctx, "GET", path, nil)
@@ -120,7 +130,7 @@ func (c *SSHKeyClient) All(ctx context.Context) ([]*SSHKey, error) {
 func (c *SSHKeyClient) AllWithOpts(ctx context.Context, opts SSHKeyListOpts) ([]*SSHKey, error) {
 	allSSHKeys := []*SSHKey{}
 
-	_, err := c.client.all(func(page int) (*Response, error) {
+	err := c.client.all(func(page int) (*Response, error) {
 		opts.Page = page
 		sshKeys, resp, err := c.List(ctx, opts)
 		if err != nil {
